#include "header.hpp"

fourvector::fourvector(){
  t = 0;
  x = 0;
  y = 0;
  z = 0;
}
fourvector::fourvector(double r0, double r1, double r2, double r3){
  t = r0;
  x = r1;
  y = r2;
  z = r3;
}

// destructor needed ???
//fourvector::~fourvector();


double fourvector::eta() const {
  static Logger logger("fourvector::eta");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  // old implementation: //  double eta = -log(tan((theta() / 2.)));
  double _eta = 0.;
  if (z != 0.){
    double eps = (pow(x, 2) + pow(y, 2)) / pow(z, 2);
    if (eps == 0.){
      if (z > 0){return +numeric_limits<double>::infinity();}
      else {return -numeric_limits<double>::infinity();}
    }
    if (eps < 1.e-13){
      if (z > 0.){_eta = 0.5 * log((r() + z) / (0.5 * z * eps));}
      else {_eta = 0.5 * log(-0.5 * z * eps / (r() - z));}
    }
    else {_eta = 0.5 * log((r() + z) / (r() - z));}
    if (munich_isnan(_eta) || abs(_eta)>1.e99) {
      logger << LOG_WARN << "computation unstable; eta=" << _eta << ", eps=" << eps << endline;
      logger << LOG_WARN << (*this) << endline;
      logger << LOG_WARN << "r=" << r() << ", z=" << z << ", old implementation: eta=" << -log(tan((theta() / 2.))) << endline_flush;
    }
  }

  //  logger << LOG_DEBUG_VERBOSE << "finished - return " << _eta << endline_flush;
  return _eta;
}

double fourvector::rapidity() const {
  //  static Logger logger("fourvector::rapidity");
  //  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;
  
  double qq = z / t;
  
  //  logger << LOG_DEBUG_VERBOSE << "finished - return " << .5 * log((qq + 1) / (1 - qq)) << endline_flush;
  return .5 * log((qq + 1) / (1 - qq));
}

// calculate distance R (definied via pseudo-rapidity) of fourvectors "fv1" and "fv2"
// replace by functions that explicitly refer to y or eta !!!
double R(const fourvector & fv1, const fourvector & fv2){
  double delta_phi, R;
  //  cosangle = (fv1.x * fv2.x + fv1.y * fv2.y + fv1.z * fv2.z) / (fv1.r() * fv2.r());
  delta_phi = abs(fv1.phi() - fv2.phi());
  if (delta_phi > pi){delta_phi = 2. * pi - delta_phi;}
  R = sqrt(pow(fv1.eta() - fv2.eta(), 2.) + pow(delta_phi, 2.));
  return R;
  //  return sqrt(pow(fv1.eta() - fv2.eta(), 2.) + pow(delta_phi, 2.));
}

double R2_eta(const fourvector & fv1, const fourvector & fv2){
  double R2_eta;
  double delta_phi;
  //  cosangle = (fv1.x * fv2.x + fv1.y * fv2.y + fv1.z * fv2.z) / (fv1.r() * fv2.r());
  delta_phi = abs(fv1.phi() - fv2.phi());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  R2_eta = pow(fv1.eta() - fv2.eta(), 2) + pow(delta_phi, 2);
  // TODO: could be optimized further
  return R2_eta;
  //  return pow(fv1.eta() - fv2.eta(), 2) + pow(delta_phi, 2);
}

double R2_rapidity(const fourvector & fv1, const fourvector & fv2){
  double R2_rapidity;
  double delta_phi;
  //  cosangle = (fv1.x * fv2.x + fv1.y * fv2.y + fv1.z * fv2.z) / (fv1.r() * fv2.r());
  delta_phi = abs(fv1.phi() - fv2.phi());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  R2_rapidity = pow(fv1.rapidity() - fv2.rapidity(), 2) + pow(delta_phi, 2);
  return R2_rapidity;
  //  return pow(fv1.rapidity() - fv2.rapidity(), 2) + pow(delta_phi, 2);
}

double R2_cosheta_cosphi(const fourvector & fv1, const fourvector & fv2){
  double R2_cosheta_cosphi;
  double delta_phi;
  //  cosangle = (fv1.x * fv2.x + fv1.y * fv2.y + fv1.z * fv2.z) / (fv1.r() * fv2.r());
  delta_phi = abs(fv1.phi() - fv2.phi());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  //  R2_cosheta_cosphi = cosh(fv1.eta() - fv2.eta()) - cos(delta_phi);
  R2_cosheta_cosphi = 2 * (cosh(fv1.eta() - fv2.eta()) - cos(delta_phi));
  return R2_cosheta_cosphi;
  //  return 2 * (cosh(fv1.eta() - fv2.eta()) - cos(delta_phi));
}

double R2_coshrapidity_cosphi(const fourvector & fv1, const fourvector & fv2){
  double R2_coshrapidity_cosphi;
  double delta_phi;
  //  cosangle = (fv1.x * fv2.x + fv1.y * fv2.y + fv1.z * fv2.z) / (fv1.r() * fv2.r());
  delta_phi = abs(fv1.phi() - fv2.phi());
  if (delta_phi > pi){delta_phi = 2 * pi - delta_phi;}
  //  R2_coshrapidity_cosphi = cosh(fv1.rapidity() - fv2.rapidity()) - cos(delta_phi);
  R2_coshrapidity_cosphi = 2 * (cosh(fv1.rapidity() - fv2.rapidity()) - cos(delta_phi));
  return R2_coshrapidity_cosphi;
  //  return 2 * (cosh(fv1.rapidity() - fv2.rapidity()) - cos(delta_phi));
}

bool fourvector::isnan(){
  if (munich_isnan(t) || munich_isnan(x) || munich_isnan(y) || munich_isnan(z)){return true;}
  else {return false;}
}

bool fourvector::isinf(){
  if (munich_isinf(t) || munich_isinf(x) || munich_isinf(y) || munich_isinf(z)){return true;}
  else {return false;}
}

bool fourvector::isinfnan(){
  if (this->isnan() || this->isinf()){return true;}
  else {return false;}
}

ostream & operator << (ostream & s, const fourvector & fv){
  s << "( " << setw(21) << setprecision(15) << fv.t << " ; " << setw(21) << setprecision(15) << fv.x << " , " << setw(21) << setprecision(15) << fv.y << " , " << setw(21) << setprecision(15) << fv.z <<" )";
  //  s << "( " << setw(14) << setprecision(8) << fv.t << " ; " << setw(14) << setprecision(8) << fv.x << " , " << setw(14) << setprecision(8) << fv.y << " , " << setw(14) << setprecision(8) << fv.z <<" )";
  return s;
}
